/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */


#include <initng.h>
#include <initng_handler.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>

#include "initng_colorprint_out.h"

#include <initng_global.h>
#include <initng_string_tools.h>
#include <initng_active_db.h>
#include <initng_plugin_hook.h>
#include <initng_static_states.h>
#include <initng_toolbox.h>

#define PE C_RED " %3i%% " C_OFF C_BLUE "%s" C_OFF MOVE_TO_R
#define P C_RED "      " C_OFF C_BLUE "%s" C_OFF MOVE_TO_R

active_db_h *lastservice;
ptype_h *last_ptype;

/* when this is set, initng wont print with cpout when system is up */
int quiet_when_up = FALSE;

static void clear_lastserv(void)
{
	if (lastservice)
		printf("\n");
	lastservice = NULL;
	last_ptype = NULL;
}


static void out_service_done(active_db_h * s)
{
	int t;

	assert(s);
	assert(s->name);

	t = MS_DIFF(s->time_current_state, s->time_last_state);

	clear_lastserv();
	if (t > 1)
	{
		if (g.sys_state == STATE_STARTING)
			printf(PE "\t[" C_GREEN "done" C_OFF "]  \t( done in %ims )\n",
				   initng_active_db_percent_started(), s->name, t);
		else
			printf(P "\t[" C_GREEN "done" C_OFF "]  \t( done in %ims )\n",
				   s->name, t);
	}
	else
	{
		if (g.sys_state == STATE_STARTING)
			printf(PE "\t[" C_GREEN "done" C_OFF "]\n",
				   initng_active_db_percent_started(), s->name);
		else
			printf(P "\t[" C_GREEN "done" C_OFF "]\n", s->name);
	}
}

static void opt_service_stop_p(active_db_h * s, const char *is)
{
	int t;

	assert(s);
	assert(s->name);

	t = MS_DIFF(s->time_current_state, s->time_last_state);

	clear_lastserv();
	if (t > 1)
	{
		if (g.sys_state == STATE_STOPPING)
			printf(PE "\t[" C_GREEN "%s" C_OFF
				   "]\t( %s in %ims )\n", initng_active_db_percent_stopped(),
				   s->name, is, is, t);
		else
			printf(P "\t[" C_GREEN "%s" C_OFF
				   "]\t( %s in %ims )\n", s->name, is, is, t);
	}
	else
	{
		if (g.sys_state == STATE_STOPPING)
			printf(PE "\t[" C_GREEN "%s" C_OFF "]\n",
				   initng_active_db_percent_stopped(), s->name, is);
		else
			printf(P "\t[" C_GREEN "%s" C_OFF "]\n", s->name, is);
	}
}


static int print_output(active_db_h * service)
{
	assert(service);
	assert(service->name);

	/* dont print hidden services */
	if (service->type && service->type->hidden == TRUE)
		return (TRUE);

	/* if quiet_when_up and system up, dont print anything */
	if (quiet_when_up && g.sys_state == STATE_UP)
		return (TRUE);

	if (IS_DOWN(service))
	{
		opt_service_stop_p(service, "stopped");
		return (TRUE);
	}

	if (IS_STARTING(service))
	{
		/* if we print this on boot, we clutter up the screen too much */
		if (g.sys_state == STATE_STARTING)
			return (TRUE);
		clear_lastserv();
		printf(P "\t[" C_GREEN "starting" C_OFF "]\n", service->name);
		return (TRUE);
	}

	if (IS_UP(service))
	{
		int t;
		process_h *process = initng_process_db_get_by_name("daemon", service);

		if (!process)
		{
			out_service_done(service);
			return (TRUE);
		}
		clear_lastserv();
		t = initng_active_db_percent_started();

		if (t > 1 && g.sys_state == STATE_STARTING)
			printf(PE "\t[" C_GREEN "started" C_OFF "]\t( pid: %i )\n",
				   t, service->name, process->pid);
		else
			printf(P "\t[" C_GREEN "started" C_OFF "]\t( pid: %i )\n",
				   service->name, process->pid);

		return (TRUE);
	}

	if (IS_STOPPING(service))
	{
		/*
		 * don't prompt that we are stopping a service, if system is shutting down, i 
		 * do think that the user is aware about this.
		 */
		if (g.sys_state == STATE_STOPPING)
			return (TRUE);

		clear_lastserv();
		printf(P "\t[" C_GREEN "stopping" C_OFF "]\n", service->name);

		return (TRUE);
	}

	/* Print all states, that is a failure state */
	if (IS_FAILED(service))
	{
		clear_lastserv();
		printf(P "\t[" C_RED "%s" C_OFF "]\n", service->name,
			   service->current_state->state_name);
	}
	return (TRUE);
}

static void print_system_state(h_sys_state state)
{

	switch (state)
	{
		case STATE_STARTING:
			clear_lastserv();
			printf("\tSystem is starting up!\n\n");
			break;
		case STATE_STOPPING:
			clear_lastserv();
			printf("\tSystem is going down!\n\n");
			break;
		case STATE_ASE:
			clear_lastserv();
			printf("\tLast process exited!\n\n");
			break;
		case STATE_EXIT:
			clear_lastserv();
			printf("\tInitng exiting!\n\n");
			break;
		case STATE_RESTART:
			clear_lastserv();
			printf("\tInitng restarting!\n\n");
			break;
		case STATE_REBOOT:
			clear_lastserv();
			printf("\n\tYour system will now REBOOT!\n");
			break;
		case STATE_HALT:
			clear_lastserv();
			printf("\n\tYour system will now HALT!\n");
			break;
		case STATE_POWEROFF:
			clear_lastserv();
			printf("\n\tYour system will now POWER_OFF!\n");
			break;
		case STATE_UP:
			{
				struct timeval now;

				/* Dont print this */
				if (quiet_when_up)
					return;

				/* get runlevel */
				active_db_h *runl = initng_active_db_find_by_name(g.runlevel);

				/* this is not critical, its youst fr runlevel up in info */
				if (!runl)
				{
					D_("Runlevel %s not found.\n", g.runlevel);
					return;
				}

				gettimeofday(&now, NULL);

				clear_lastserv();
				printf("\n\n\trunlevel \"%s\" up in, %ims.\n\n",
					   g.runlevel, MS_DIFF(now, runl->last_rought_time));

				break;
			}
		default:
			break;
	}
	D_("print_system_state(): new system state: %i\n", state);
}

static int print_program_output(active_db_h * service, process_h * x,
								char *buffer_pos)
{
	/*
	   TODO here:
	   we should have an internal list of services and "our" current position in them.

	   That way when this function is called we can print every full line from plugin_pos.
	   This way fsck will look nice, along with an "internal" database of write positions we can cache data so we print every 5 seconds or on int forceflush.
	 */
	int i = 0;

	assert(service);
	assert(service->name);
	assert(x);
	S_;

	/* if quiet_when_up and system up, dont print anything */
	if (quiet_when_up && g.sys_state == STATE_UP)
		return (TRUE);

	D_(" from service \"%s\"\n", service->name);
	/*
	   printf("buffer_pos: %i\n", buffer_pos);
	   printf("datalen: %i\n", datalen);
	   printf("Buffer: \n################\n%s\n##########\n\n",x->buffer);
	 */
	/* a first while loop that sorts out crap */
	while (buffer_pos[i] != '\0')
	{
		/*  remove lines with " [2]  Done " that bash generates. */
		if (buffer_pos[i] == '[' && buffer_pos[i + 2] == ']')
		{
			/* jump to next line */
			while (buffer_pos[i] && buffer_pos[i] != '\n')
				i++;
		}

		/* if there are stupid tokens, go to next char, and run while again. */
		if (buffer_pos[i] == ' ' || buffer_pos[i] == '\n'
			|| buffer_pos[i] == '\t')
		{
			i++;
			continue;
		}

		/* else break */
		break;
	}

	/* Make sure that there is anything left to write */
	if (strlen(&buffer_pos[i]) < 2)
	{
		/* its okay anyway */
		return (TRUE);
	}

	if (lastservice != service && last_ptype != x->pt)
	{
		clear_lastserv();
		printf("\n" C_CYAN " %s %s:" C_OFF, service->name, x->pt->name);
		/* print our special indented newline */
		putchar('\n');
		putchar(' ');
		putchar(' ');
		lastservice = service;
		last_ptype = x->pt;
	}
	else
	{
		D_("Lastservice == service, won't print header.\n");
	}



	/* while buffer lasts */
	while (buffer_pos[i] != '\0')
	{
		/*  remove lines with " [2]  Done " that bash generates. */
		if (buffer_pos[i] == '[' && buffer_pos[i + 2] == ']')
		{
			while (buffer_pos[i] && buffer_pos[i] != '\n')
				i++;
		}

		/* if this are a newline */
		if (buffer_pos[i] == '\n')
		{
			/* print our special indented newline instead */
			putchar('\n');
			putchar(' ');
			putchar(' ');
			i++;
			/* skip spaces, on newline. */
			while (buffer_pos[i]
				   && (buffer_pos[i] == ' ' || buffer_pos[i] == '\t'))
				i++;
			continue;
		}

		/* ok, now put the char, and go to next. */
		putchar(buffer_pos[i]);
		i++;
	}

	/* flush any buffered output to the screen */
	fflush(stdout);
	return (TRUE);
}


static int cp_print_error(e_mt mt, const char *file, const char *func,
						  int line, const char *format, va_list arg)
{
	struct tm *ts;
	time_t t;

	switch (mt)
	{
		case MSG_FAIL:
		case MSG_WARN:
			t = time(0);
			ts = localtime(&t);
#ifdef DEBUG
			fprintf(stderr, "\n\n ** \"%s\", %s()  line:%i:\n", file, func,
					line);
#endif
			fprintf(stderr, " %.2i:%.2i:%.2i -- %s:\t", ts->tm_hour,
					ts->tm_min, ts->tm_sec, mt == MSG_FAIL ? "FAIL" : "WARN");
			vfprintf(stderr, format, arg);
			break;
		default:
			vfprintf(stderr, format, arg);
			break;
	}
	return (TRUE);
}

int module_init(int api_version)
{
	int i;

	if (api_version != API_VERSION)
	{
		F_("This module is compiled for api_version %i version and initng is compiled on %i version, won't load this module!\n", API_VERSION, api_version);
		return (FALSE);
	}

	/* only add this module if this is a real init */
	if (g.i_am != I_AM_INIT && g.i_am != I_AM_FAKE_INIT)
		return (TRUE);


	for (i = 0; g.Argv[i]; i++)
	{
		if (strstr(g.Argv[i], "quiet_when_up"))
			quiet_when_up = TRUE;
	}


	printf("\n" C_BLUE "\tNext Generation Init version ( %s )" C_OFF,
		   INITNG_VERSION);
	printf("\n" C_GREEN "\thttp://www.initng.org\n" C_OFF);
	printf("\tAuthor: Jimmy Wennlund <jimmy.wennlund@gmail.com>\n");
	printf("\tIf you find initng useful, please consider a small donation.\n\n");
	D_("module_init();\n");
	lastservice = NULL;
	initng_plugin_hook_register(&g.ERR_MSG, 10, &cp_print_error);
	initng_plugin_hook_register(&g.IS_CHANGE, 80, &print_output);
	initng_plugin_hook_register(&g.SWATCHERS, 80, &print_system_state);
	initng_plugin_hook_register(&g.PIPEWATCHERS, 50, &print_program_output);
	return (TRUE);
}

void module_unload(void)
{
	D_("color_out: module_unload();\n");
	if (g.i_am != I_AM_INIT && g.i_am != I_AM_FAKE_INIT)
		return;


	initng_plugin_hook_unregister(&g.IS_CHANGE, &print_output);
	initng_plugin_hook_unregister(&g.SWATCHERS, &print_system_state);
	initng_plugin_hook_unregister(&g.PIPEWATCHERS, &print_program_output);
	initng_plugin_hook_unregister(&g.ERR_MSG, &cp_print_error);
	printf("  Goodbye\n");
}
